% Image denoising using adaptative wavelet shrinkage based on noise model

% Output:   den_img - denoised image (2D array)

% Input:    raw_img - raw (noisy) image (2D array)
%           wname - name (string) of the wavelet used for image decomposition (default: 'bior3.7')
%           decmp_depth - number of wavelet decomposition levels used for denoising (2 or 3)
%           cheb_fit_coeff - coefficients of wavelet noise model (Chebyshev polynomial). Loaded from cheb_fit_coeff.mat.             
%           poissn_min - minimum allowed Poisson noise coefficient (default 0)
%           poissn_max - maximum allowed Poisson noise coefficient (default 2.25)
%           addn_min - minimum allowed additive noise coefficient (gaussian sigma, default 10)
%           addn_max - maximum allowed additive noise coefficient (gaussian sigma, default 60)
%           poissn - Poisson noise coefficient for the raw_img
%           addn - additive noise coefficient for the raw_img 
%           conf_lev - confidence level of model-based noise estimation (default 1.0, increases denoising strength)
            
function den_img=model_den(raw_img,wname,decmp_depth, cheb_fit_coeff,poissn_min,poissn_max,addn_min,addn_max,poissn,addn,conf_lev)


%convert image

raw_img=double(raw_img);

%create wavelet filters

[Lo_D,Hi_D,Lo_R,Hi_R] = wfilters(wname);
Lo_D_Sq=Lo_D.*Lo_D;
Hi_D_Sq=Hi_D.*Hi_D;


%caclucate position of decomposition level data in the data vector 

pointer_start=0;
pointer_end=0;
index=1;

%decompose noise image

[CSig_nois,SSig_nois] = wavedec2(raw_img,decmp_depth,Lo_D,Hi_D);

%calculate data pointers

for k=1:(size(SSig_nois,1)-1)

    for l=1:3

        if ((k>1) || (l==1))
            pointer_start=pointer_end+1;
            pointer_end=(SSig_nois(k,1)^2)+pointer_start-1;

            wdata_pointer(index,1)=pointer_start;
            wdata_pointer(index,2)=pointer_end;
            wdata_pointer(index,3)=SSig_nois(k,1);    

            if ((k==1) && (l==1))
                wdata_pointer(index,4)=0;
            else    
                wdata_pointer(index,4)=l;
            end

            index=index+1;
        end
    end

end


%Soft thresholding using noise model
        
%decompose noise image 


for m=0:(decmp_depth-2)

    ind_LH=size(wdata_pointer,1)-(m*3+2);
    ind_HL=size(wdata_pointer,1)-(m*3+1);
    ind_HH=size(wdata_pointer,1)-m*3;    

    %caclulate child/parent noise fit coefficients

    cheb_coeff_table=cheb_fit_coeff{1,3-m};
    coeff_vec_LH(1,1)=chebyshev5(cheb_coeff_table(1,:),poissn_min,poissn_max,addn_min,addn_max,poissn,addn);
    coeff_vec_LH(1,2)=chebyshev5(cheb_coeff_table(2,:),poissn_min,poissn_max,addn_min,addn_max,poissn,addn);
    coeff_vec_LH(1,3)=chebyshev5(cheb_coeff_table(3,:),poissn_min,poissn_max,addn_min,addn_max,poissn,addn);

    cheb_coeff_table=cheb_fit_coeff{2,3-m};
    coeff_vec_HL(1,1)=chebyshev5(cheb_coeff_table(1,:),poissn_min,poissn_max,addn_min,addn_max,poissn,addn);
    coeff_vec_HL(1,2)=chebyshev5(cheb_coeff_table(2,:),poissn_min,poissn_max,addn_min,addn_max,poissn,addn);
    coeff_vec_HL(1,3)=chebyshev5(cheb_coeff_table(3,:),poissn_min,poissn_max,addn_min,addn_max,poissn,addn);

    cheb_coeff_table=cheb_fit_coeff{3,3-m};
    coeff_vec_HH(1,1)=chebyshev5(cheb_coeff_table(1,:),poissn_min,poissn_max,addn_min,addn_max,poissn,addn);
    coeff_vec_HH(1,2)=chebyshev5(cheb_coeff_table(2,:),poissn_min,poissn_max,addn_min,addn_max,poissn,addn);
    coeff_vec_HH(1,3)=chebyshev5(cheb_coeff_table(3,:),poissn_min,poissn_max,addn_min,addn_max,poissn,addn); 
    
    
    %create filters for local variance and average calculation

    variance_neighb=ones(3,3);

    %calculate child wavelet coefficients and their variance

    CSig_nois_LH=CSig_nois(wdata_pointer(ind_LH,1):wdata_pointer(ind_LH,2));
    CSig_nois_HL=CSig_nois(wdata_pointer(ind_HL,1):wdata_pointer(ind_HL,2));
    CSig_nois_HH=CSig_nois(wdata_pointer(ind_HH,1):wdata_pointer(ind_HH,2));  

    %calculate parent wavelet coefficients and their variance

    CSig_nois_par_LH=reshape(CSig_nois(wdata_pointer(ind_LH-3,1):wdata_pointer(ind_LH-3,2)),wdata_pointer(ind_LH-3,3),wdata_pointer(ind_LH-3,3));
    CSig_nois_par_LH=imresize(abs(CSig_nois_par_LH), [wdata_pointer(ind_LH,3) wdata_pointer(ind_LH,3)],'bicubic'); 
    CSig_nois_par_LH=reshape(CSig_nois_par_LH, 1, wdata_pointer(ind_LH,3)^2);

    CSig_nois_par_HL=reshape(CSig_nois(wdata_pointer(ind_HL-3,1):wdata_pointer(ind_HL-3,2)),wdata_pointer(ind_HL-3,3),wdata_pointer(ind_HL-3,3));
    CSig_nois_par_HL=imresize(abs(CSig_nois_par_HL), [wdata_pointer(ind_HL,3) wdata_pointer(ind_HL,3)],'bicubic'); 
    CSig_nois_par_HL=reshape(CSig_nois_par_HL, 1, wdata_pointer(ind_HL,3)^2);

    CSig_nois_par_HH=reshape(CSig_nois(wdata_pointer(ind_HH-3,1):wdata_pointer(ind_HH-3,2)),wdata_pointer(ind_HH-3,3),wdata_pointer(ind_HH-3,3));
    CSig_nois_par_HH=imresize(abs(CSig_nois_par_HH), [wdata_pointer(ind_HH,3) wdata_pointer(ind_HH,3)],'bicubic'); 
    CSig_nois_par_HH=reshape(CSig_nois_par_HH, 1, wdata_pointer(ind_HH,3)^2);
    
    %perform wavelet shrinkage

    wav_noise_LH=(coeff_vec_LH(1)+abs(CSig_nois_LH)*coeff_vec_LH(2)+abs(CSig_nois_par_LH)*coeff_vec_LH(3))*conf_lev;
    CSig_nois_LH=max((abs(CSig_nois_LH)-abs(wav_noise_LH)),0).*sign(CSig_nois_LH); 
    CSig_nois(wdata_pointer(ind_LH,1):wdata_pointer(ind_LH,2))=CSig_nois_LH;

    wav_noise_HL=(coeff_vec_HL(1)+abs(CSig_nois_HL)*coeff_vec_HL(2)+abs(CSig_nois_par_HL)*coeff_vec_HL(3))*conf_lev;
    CSig_nois_HL=max((abs(CSig_nois_HL)-abs(wav_noise_HL)),0).*sign(CSig_nois_HL); 
    CSig_nois(wdata_pointer(ind_HL,1):wdata_pointer(ind_HL,2))=CSig_nois_HL;    

    wav_noise_HH=(coeff_vec_HH(1)+abs(CSig_nois_HH)*coeff_vec_HH(2)+abs(CSig_nois_par_HH)*coeff_vec_HH(3))*conf_lev;
    CSig_nois_HH=max((abs(CSig_nois_HH)-abs(wav_noise_HH)),0).*sign(CSig_nois_HH);
    CSig_nois(wdata_pointer(ind_HH,1):wdata_pointer(ind_HH,2))=CSig_nois_HH; 


end 

%caclulate  noise model MSE

den_img = waverec2(CSig_nois,SSig_nois,Lo_R,Hi_R);


% calculation ot Chebyshev 2D polynomial of 5th order 

function chval=chebyshev5(coeff_vec,x_min,x_max,y_min,y_max,x,y)

sc_val_vec=zeros(1,size(coeff_vec,2));

%calculate scaled values

x=(x-x_min)/(x_max-x_min)*2-1;
y=(y-y_min)/(y_max-y_min)*2-1;

sc_val_vec(1)=1;
sc_val_vec(2)=cos(acos(x));
sc_val_vec(3)=cos(acos(y));
sc_val_vec(4)=cos(2*acos(x));
sc_val_vec(5)=cos(acos(x))*cos(acos(y));
sc_val_vec(6)=cos(2*acos(y));
sc_val_vec(7)=cos(3*acos(x));
sc_val_vec(8)=cos(2*acos(x))*cos(acos(y));
sc_val_vec(9)=cos(acos(x))*cos(2*acos(y));
sc_val_vec(10)=cos(3*acos(y));
sc_val_vec(11)=cos(4*acos(x));
sc_val_vec(12)=cos(3*acos(x))*cos(acos(y));
sc_val_vec(13)=cos(2*acos(y))*cos(2*acos(x));
sc_val_vec(14)=cos(acos(x))*cos(3*acos(y));
sc_val_vec(15)=cos(4*acos(y));
sc_val_vec(16)=cos(5*acos(x));
sc_val_vec(17)=cos(4*acos(x))*cos(acos(y));
sc_val_vec(18)=cos(3*acos(x))*cos(2*acos(y));
sc_val_vec(19)=cos(2*acos(x))*cos(3*acos(y));
sc_val_vec(20)=cos(acos(x))*cos(4*acos(y));
sc_val_vec(21)=cos(5*acos(y));

chval=sum(coeff_vec.*sc_val_vec);



